from ase import Atoms
from icet import ClusterSpace
from icet.core.structure import Structure
from mchammer.observers.base_observer import BaseObserver
from typing import List, Dict
import numpy as np


class SiteOccupancyObserver(BaseObserver):
    """
    This class represents a site occupation factor (SOF) observer.

    A SOF observer allows to compute the site occupation factors along the
    trajectory sampled by a Monte Carlo (MC) simulation.

    Parameters
    ----------
    cluster_space : icet.ClusterSpace
        cluster space from which the allowed species are extracted

    sites : dict(str, list(int))
        dictionary containing lists of sites that are to be considered,
        which keys will be taken as the names of the sites

    supercell : ase.Atoms
        a typical supercell, which is used to determine the allowed species

    interval : int
        observation interval during the Monte Carlo simulation

    Attributes
    ----------
    tag : str
        name of observer

    interval : int
        observation interval

    Example
    -------
    The following snippet illustrate how to use the site occupancy factor (SOF)
    observer in a Monte Carlo simulation of a surface slab. Here, the SOF
    observer is used to monitor the concentrations of different species at the
    surface, the first subsurface layer, and the remaining "bulk". A minimal
    cluster expansion is used with slightly modified surface interactions in
    order to obtain an example that can be run without much ado. In practice,
    one should of course use a proper cluster expansion::

        from ase.build import fcc111
        from icet import ClusterExpansion, ClusterSpace
        from mchammer.calculators import ClusterExpansionCalculator
        from mchammer.ensembles import CanonicalEnsemble
        from mchammer.observers import SiteOccupancyObserver

        # prepare reference structure
        prim = fcc111('Au', size=(1, 1, 10), vacuum=10.0)
        prim.translate((0.1, 0.1, 0.0))
        prim.wrap()
        prim.pbc = True  # icet requires pbc in all directions

        # prepare cluster expansion
        cs = ClusterSpace(prim, cutoffs=[3.7], chemical_symbols=['Ag', 'Au'])
        params = [0] + 5 * [0] + 10 * [0.1]
        params[1] = 0.01
        params[6] = 0.12
        ce = ClusterExpansion(cs, params)
        print(ce)

        # prepare initial configuration based on a 2x2 supercell
        structure = prim.repeat((2, 2, 1))
        for k in range(20):
            structure[k].symbol = 'Ag'

        # set up MC simulation
        calc = ClusterExpansionCalculator(structure, ce)
        mc = CanonicalEnsemble(structure=structure, calculator=calc, temperature=600,
                               data_container='myrun_sof.dc')

        # set up observer and attach it to the MC simulation
        sites = {'surface': [0, 9], 'subsurface': [1, 8],
                 'bulk': list(range(2, 8))}
        sof = SiteOccupancyObserver(cs, sites, structure, interval=len(structure))
        mc.attach_observer(sof)

        # run 1000 trial steps
        mc.run(1000)

    After having run this snippet one can access the SOFs via the data
    container::

        print(mc.data_container.data)
    """

    def __init__(self, cluster_space: ClusterSpace,
                 sites: Dict[str, List[int]],
                 supercell: Atoms,
                 interval: int = None) -> None:
        super().__init__(interval=interval, return_type=dict,
                         tag='SiteOccupancyObserver')

        self._sites = {site: sorted(indices)
                       for site, indices in sites.items()}

        self._set_allowed_species(cluster_space, supercell)

    def _set_allowed_species(self,
                             cluster_space: ClusterSpace,
                             supercell: Atoms):
        """
        Set the allowed species for the selected sites in the Atoms object

        Parameters
        ----------
        cluster_space
            Cluster space implicitly defining allowed species
        supercell
            Specific supercell (consistent with cluster_space) whose
            allowed species are to be determined
        """

        primitive_structure = Structure.from_atoms(
            cluster_space.primitive_structure)
        chemical_symbols = cluster_space.get_chemical_symbols()

        if len(chemical_symbols) == 1:
            # If the allowed species are the same for all sites no loop is
            # required
            allowed_species = {site: chemical_symbols[0] for
                               site in self._sites.keys()}
        else:
            # Loop over the lattice sites to find the allowed species
            allowed_species = {}
            for site, indices in self._sites.items():
                allowed_species[site] = None
                positions = supercell.get_positions()[np.array(indices)]
                lattice_sites =\
                    primitive_structure.find_lattice_sites_by_positions(
                        positions)
                for l, lattice_site in enumerate(lattice_sites):
                    species = chemical_symbols[lattice_site.index]
                    # check that the allowed species are equal for all sites
                    if allowed_species[site] is not None and\
                            species != allowed_species[site]:
                        raise Exception("The allowed species {} for the site"
                                        " with index {} differs from the"
                                        " result {} for the previous index"
                                        " ({})!".format(species, indices[l],
                                                        allowed_species[site],
                                                        indices[l-1]))
                    allowed_species[site] = species

        self._allowed_species = allowed_species

    def get_observable(self, structure: Atoms) -> Dict[str, List[float]]:
        """
        Returns the site occupation factors for a given atomic configuration.

        Parameters
        ----------
        structure
            input atomic structure.
        """

        chemical_symbols = np.array(structure.get_chemical_symbols())
        sofs = {}
        for site, indices in self._sites.items():
            counts = {species: 0 for species in self._allowed_species[site]}
            symbols, sym_counts = np.unique(chemical_symbols[indices],
                                            return_counts=True)
            for sym, count in zip(symbols, sym_counts):
                counts[sym] += count

            for species in counts.keys():
                key = 'sof_{}_{}'.format(site, species)
                sofs[key] = float(counts[species]) / len(indices)

        return sofs



from collections import namedtuple
from typing import Dict

import numpy as np

from ase import Atoms
from icet import ClusterSpace
from mchammer.observers import ClusterCountObserver
from mchammer.observers.base_observer import BaseObserver

ClusterCountInfo = namedtuple('ClusterCountInfo', ['counts', 'dc_tags'])


class BinaryShortRangeOrderObserver(BaseObserver):
    """
    This class represents a short range order (SRO) observer for a
    binary system.


    Parameters
    ----------
    cluster_space : icet.ClusterSpace
        cluster space used for initialization
    structure : ase.Atoms
        defines the lattice which the observer will work on
    interval : int
        observation interval during the Monte Carlo simulation
    radius : float
        the maximum radius  for the neigbhor shells considered

    Attributes
    ----------
    tag : str
        human readable observer name (`BinaryShortRangeOrderObserver`)
    interval : int
        observation interval

    Example
    -------
    The following snippet illustrate how to use the short-range order (SRO)
    observer in a Monte Carlo simulation of a bulk supercell. Here, the
    parameters of the cluster expansion are set to emulate a simple Ising model
    in order to obtain an example that can be run without modification. In
    practice, one should of course use a proper cluster expansion::

        from ase.build import bulk
        from icet import ClusterExpansion, ClusterSpace
        from mchammer.calculators import ClusterExpansionCalculator
        from mchammer.ensembles import CanonicalEnsemble
        from mchammer.observers import BinaryShortRangeOrderObserver

        # prepare cluster expansion
        # the setup emulates a second nearest-neighbor (NN) Ising model
        # (zerolet and singlet ECIs are zero; only first and second neighbor
        # pairs are included)
        prim = bulk('Au')
        cs = ClusterSpace(prim, cutoffs=[4.3], chemical_symbols=['Ag', 'Au'])
        ce = ClusterExpansion(cs, [0, 0, 0.1, -0.02])

        # prepare initial configuration
        nAg = 10
        structure = prim.repeat(3)
        structure.set_chemical_symbols(nAg * ['Ag'] + (len(structure) - nAg) * ['Au'])

        # set up MC simulation
        calc = ClusterExpansionCalculator(structure, ce)
        mc = CanonicalEnsemble(structure=structure, calculator=calc, temperature=600,
                               data_container='myrun_sro.dc')

        # set up observer and attach it to the MC simulation
        sro = BinaryShortRangeOrderObserver(cs, structure, interval=len(structure),
                                            radius=4.3)
        mc.attach_observer(sro)

        # run 1000 trial steps
        mc.run(1000)

    After having run this snippet one can access the SRO parameters via the
    data container::

        print(mc.data_container.data)
    """

    def __init__(self, cluster_space, structure: Atoms,
                 interval: int, radius: float) -> None:
        super().__init__(interval=interval, return_type=dict,
                         tag='BinaryShortRangeOrderObserver')

        self._structure = structure

        self._cluster_space = ClusterSpace(
            structure=cluster_space.primitive_structure,
            cutoffs=[radius],
            chemical_symbols=cluster_space.chemical_symbols)
        self._cluster_count_observer = ClusterCountObserver(
            cluster_space=self._cluster_space, structure=structure,
            interval=interval)

        self._sublattices = self._cluster_space.get_sublattices(structure)
        binary_sublattice_counts = 0
        for symbols in self._sublattices.allowed_species:
            if len(symbols) == 2:
                binary_sublattice_counts += 1
                self._symbols = sorted(symbols)
            elif len(symbols) > 2:
                raise ValueError('Cluster space has more than two allowed'
                                 ' species on a sublattice. '
                                 'Allowed species: {}'.format(symbols))
        if binary_sublattice_counts != 1:
            raise ValueError('Number of binary sublattices must equal one,'
                             ' not {}'.format(binary_sublattice_counts))

    def get_observable(self, structure: Atoms) -> Dict[str, float]:
        """Returns the value of the property from a cluster expansion
        model for a given atomic configurations.

        Parameters
        ----------
        structure
            input atomic structure
        """

        self._cluster_count_observer._generate_counts(structure)
        df = self._cluster_count_observer.count_frame

        symbol_counts = self._get_atom_count(structure)
        conc_B = self._get_concentrations(structure)[self._symbols[0]]

        pair_orbit_indices = set(
            df.loc[df['order'] == 2]['orbit_index'].tolist())
        N = symbol_counts[self._symbols[0]] + symbol_counts[self._symbols[1]]
        sro_parameters = {}
        for k, orbit_index in enumerate(sorted(pair_orbit_indices)):
            orbit_df = df.loc[df['orbit_index'] == orbit_index]
            A_B_pair_count = 0
            total_count = 0
            total_A_count = 0
            for i, row in orbit_df.iterrows():
                total_count += row.cluster_count
                if self._symbols[0] in row.occupation:
                    total_A_count += row.cluster_count
                if self._symbols[0] in row.occupation and \
                        self._symbols[1] in row.occupation:
                    A_B_pair_count += row.cluster_count

            key = 'sro_{}_{}'.format(self._symbols[0], k+1)
            Z_tot = symbol_counts[self._symbols[0]] * 2 * total_count / N
            if conc_B == 1 or Z_tot == 0:
                value = 0
            else:
                value = 1 - A_B_pair_count/(Z_tot * (1-conc_B))
            sro_parameters[key] = value

        return sro_parameters

    def _get_concentrations(self, structure: Atoms) -> Dict[str, float]:
        """Returns concentrations for each species relative its
        sublattice.

        Parameters
        ----------
        structure
            the configuration that will be analyzed
        """
        occupation = np.array(structure.get_chemical_symbols())
        concentrations = {}
        for sublattice in self._sublattices:
            if len(sublattice.chemical_symbols) == 1:
                continue
            for symbol in sublattice.chemical_symbols:
                symbol_count = occupation[sublattice.indices].tolist().count(
                    symbol)
                concentration = symbol_count / len(sublattice.indices)
                concentrations[symbol] = concentration
        return concentrations

    def _get_atom_count(self, structure: Atoms) -> Dict[str, float]:
        """Returns atom counts for each species relative its
        sublattice.

        Parameters
        ----------
        structure
            the configuration that will be analyzed
        """
        occupation = np.array(structure.get_chemical_symbols())
        counts = {}
        for sublattice in self._sublattices:
            if len(sublattice.chemical_symbols) == 1:
                continue
            for symbol in sublattice.chemical_symbols:
                symbol_count = occupation[sublattice.indices].tolist().count(
                    symbol)
                counts[symbol] = symbol_count
        return counts



"""
Handling of Hermite Normal Form matrices
"""

import numpy as np
from .smith_normal_form import SmithNormalForm


class HermiteNormalForm(object):
    """
    Hermite Normal Form matrix.
    """

    def __init__(self, H, rotations, translations, basis_shifts):
        self.H = H
        self.snf = SmithNormalForm(H)
        self.transformations = []
        self.compute_transformations(rotations, translations, basis_shifts)

    def compute_transformations(self, rotations, translations, basis_shifts,
                                tol=1e-3):
        """
        Save transformations (based on rotations) that turns the supercell
        into an equivalent supercell. Precompute these transformations,
        consisting of permutation as well as translation and basis shift, for
        later use.

        Parameters
        ----------
        rotations : list of ndarrays
        translations : list of ndarrays
        basis_shifts : ndarray
        """

        for R, T, basis_shift in zip(rotations, translations,
                                     basis_shifts):
            check = np.dot(np.dot(np.linalg.inv(self.H), R), self.H)
            check = check - np.round(check)
            if (abs(check) < tol).all():
                LRL = np.dot(self.snf.L, np.dot(R, np.linalg.inv(self.snf.L)))

                # Should be an integer matrix
                assert (abs(LRL - np.round(LRL)) < tol).all()
                LRL = np.round(LRL).astype(np.int64)
                LT = np.dot(T, self.snf.L.T)
                self.transformations.append([LRL, LT, basis_shift])


def yield_hermite_normal_forms(det, pbc):
    """
    Yield all Hermite Normal Form matrices with determinant det.

    Parameters
    ----------
    det : int
        Target determinant of HNFs
    pbc : list of bools
        Periodic boundary conditions of the primitive structure

    Yields
    ------
    ndarray
        3x3 HNF matrix
    """

    # 1D
    if sum(pbc) == 1:
        hnf = np.eye(3, dtype=int)
        for i, bc in enumerate(pbc):
            if bc:
                hnf[i, i] = det
                break
        yield hnf

    # 2D
    elif sum(pbc) == 2:
        for a in range(1, det + 1):
            if det % a == 0:
                c = det // a
                for b in range(0, c):
                    if not pbc[0]:
                        hnf = [[1, 0, 0],
                               [0, a, 0],
                               [0, b, c]]
                    elif not pbc[1]:
                        hnf = [[a, 0, 0],
                               [0, 1, 0],
                               [b, 0, c]]
                    else:
                        hnf = [[a, 0, 0],
                               [b, c, 0],
                               [0, 0, 1]]
                    yield np.array(hnf)

    # 3D
    else:
        for a in range(1, det + 1):
            if det % a == 0:
                for c in range(1, det // a + 1):
                    if det // a % c == 0:
                        f = det // (a * c)
                        for b in range(0, c):
                            for d in range(0, f):
                                for e in range(0, f):
                                    hnf = [[a, 0, 0],
                                           [b, c, 0],
                                           [d, e, f]]
                                    yield np.array(hnf)


def yield_reduced_hnfs(ncells, symmetries, pbc, tol=1e-3):
    """
    For a fixed determinant N (i.e., a number of atoms N), yield all
    Hermite Normal Forms (HNF) that are inequivalent under symmetry
    operations of the parent lattice.'

    Parameters
    ----------
    N : int
        Determinant (or, equivalently, the number of atoms) of the HNF.
    symmetries : dict of lists
        Symmetry operations of the parent lattice.
    pbc : list of bools
        Periodic boundary conditions of the primitive structure

    Yields
    ------
    list of ndarrays
        Symmetrically inequivalent HNFs with determinant N.
    """
    rotations = symmetries['rotations']
    translations = symmetries['translations']
    basis_shifts = symmetries['basis_shifts']
    hnfs = []

    for hnf in yield_hermite_normal_forms(ncells, pbc):

        # Throw away HNF:s that yield equivalent supercells
        hnf_inv = np.linalg.inv(hnf)
        duplicate = False
        for R in rotations:
            HR = np.dot(hnf_inv, R)
            for hnf_previous in hnfs:
                check = np.dot(HR, hnf_previous.H)
                check = check - np.round(check)
                if (abs(check) < tol).all():
                    duplicate = True
                    break
            if duplicate:
                break
        if duplicate:
            continue

        # If it's not a duplicate, save the hnf
        # and the supercell so that it can be compared to
        hnf = HermiteNormalForm(hnf, rotations, translations, basis_shifts)
        hnfs.append(hnf)
        yield hnf



import random
from numpy import array
from typing import Dict, List, Tuple
from ase import Atoms
from icet.core.sublattices import Sublattices
from icet.tools.geometry import atomic_number_to_chemical_symbol


class SwapNotPossibleError(Exception):
    pass


class ConfigurationManager(object):
    """
    The ConfigurationManager owns and handles information pertaining to a
    configuration being sampled in a Monte Carlo simulation.

    Parameters
    ----------
    structure : ase.Atoms
        configuration to be handled
    sublattices : :class:`Sublattices <icet.core.sublattices.Sublattices>`
        sublattices class used to define allowed occupations and so on

    Todo
    ----
    * revise docstrings
    """

    def __init__(self, structure: Atoms, sublattices: Sublattices) -> None:

        self._structure = structure.copy()
        self._occupations = self._structure.numbers
        self._sublattices = sublattices

        self._sites_by_species = self._get_sites_by_species()

    def _get_sites_by_species(self) -> List[Dict[int, List[int]]]:
        """Returns the sites that are occupied for each species.  Each
        dictionary represents one sublattice where the key is the
        species (by atomic number) and the value is the list of sites
        occupied by said species in the respective sublattice.
        """
        sites_by_species = []
        for sl in self._sublattices:
            species_dict = {key: [] for key in sl.atomic_numbers}
            for site in sl.indices:
                species_dict[self._occupations[site]].append(site)
            sites_by_species.append(species_dict)
        return sites_by_species

    @property
    def occupations(self) -> List[int]:
        """ occupation vector of the configuration (copy) """
        return self._occupations.copy()

    @property
    def sublattices(self) -> Sublattices:
        """sublattices of the configuration"""
        return self._sublattices

    @property
    def structure(self) -> Atoms:
        """ atomic structure associated with configuration (copy) """
        structure = self._structure.copy()
        structure.set_atomic_numbers(self.occupations)
        return structure

    def get_occupations_on_sublattice(self, sublattice_index: int) -> List[int]:
        """
        Returns the occupations on one sublattice.

        Parameters
        ---------
        sublattice_index
            the sublattice for which the occupations should be returned
        """
        sl = self.sublattices[sublattice_index]
        return list(self.occupations[sl.indices])

    def is_swap_possible(self, sublattice_index: int,
                         allowed_species: List[int] = None) -> bool:
        """ Checks if swap is possible on specific sublattice.

        Parameters
        ----------
        sublattice_index
            index of sublattice to be checked
        allowed_species
            list of atomic numbers for allowed species
         """
        sl = self.sublattices[sublattice_index]
        if allowed_species is None:
            swap_symbols = set(self.occupations[sl.indices])
        else:
            swap_symbols = set([o for o in self.occupations[sl.indices] if o in
                                allowed_species])
        return len(swap_symbols) > 1

    def get_swapped_state(self, sublattice_index: int,
                          allowed_species: List[int] = None
                          ) -> Tuple[List[int], List[int]]:
        """Returns two random sites (first element of tuple) and their
        occupation after a swap (second element of tuple).  The new
        configuration will obey the occupation constraints associated
        with the configuration mananger.

        Parameters
        ----------
        sublattice_index
            sublattice from which to pick sites
        allowed_species
            list of atomic numbers for allowed species
        """
        # pick the first site
        if allowed_species is None:
            available_sites =\
                self.sublattices[sublattice_index].indices
        else:
            available_sites =\
                [s for Z in allowed_species for s in
                 self._get_sites_by_species()[sublattice_index][Z]]

        try:
            site1 = random.choice(available_sites)
        except IndexError:
            raise SwapNotPossibleError(
                'Sublattice {} is empty.'.format(sublattice_index))

        # pick the second site
        if allowed_species is None:
            possible_swap_species = \
                set(self._sublattices.get_allowed_numbers_on_site(site1)) - \
                set([self._occupations[site1]])
        else:
            possible_swap_species = \
                set(allowed_species) - set([self._occupations[site1]])
        possible_swap_sites = []
        for Z in possible_swap_species:
            possible_swap_sites.extend(self._sites_by_species[sublattice_index][Z])

        possible_swap_sites = array(possible_swap_sites)

        try:
            site2 = random.choice(possible_swap_sites)
        except IndexError:
            raise SwapNotPossibleError(
                'Cannot swap on sublattice {} since it is full of {} species .'
                .format(sublattice_index,
                        atomic_number_to_chemical_symbol([self._occupations[site1]])[0]))

        return ([site1, site2], [self._occupations[site2], self._occupations[site1]])

    def get_flip_state(self, sublattice_index: int,
                       allowed_species: List[int] = None) -> Tuple[int, int]:
        """
        Returns a site index and a new species for the site.

        Parameters
        ----------
        sublattice_index
            index of sublattice from which to pick a site
        allowed_species
            list of atomic numbers for allowed species
        """
        if allowed_species is None:
            available_sites = self._sublattices[sublattice_index].indices
        else:
            available_sites = [s for Z in allowed_species for s in
                               self._get_sites_by_species()[sublattice_index][Z]]

        site = random.choice(available_sites)
        if allowed_species is not None:
            species = random.choice(list(
                set(allowed_species) - set([self._occupations[site]])))
        else:
            species = random.choice(list(
                set(self._sublattices[sublattice_index].atomic_numbers) -
                set([self._occupations[site]])))
        return site, species

    def update_occupations(self, sites: List[int], species: List[int]):
        """
        Updates the occupation vector of the configuration being sampled.
        This will change the state in both the configuration in the calculator
        and the configuration manager.

        Parameters
        ----------
        sites
            indices of sites of the configuration to change
        species
            new occupations by atomic number
        """

        # Update sublattices
        for site, new_Z in zip(sites, species):
            if 0 > new_Z > 118:
                raise ValueError('Invalid new species {} on site {}'.format(new_Z, site))
            if len(self._occupations) >= site < 0:
                raise ValueError('Site {} is not a valid site index'.format(site))
            old_Z = self._occupations[site]
            sublattice_index = self.sublattices.get_sublattice_index(site)

            if new_Z not in self.sublattices[sublattice_index].atomic_numbers:
                raise ValueError('Invalid new species {} on site {}'.format(new_Z, site))

            # Remove site from list of sites for old species
            self._sites_by_species[sublattice_index][old_Z].remove(site)
            # Add site to list of sites for new species
            try:
                self._sites_by_species[sublattice_index][new_Z].append(site)
            except KeyError:
                raise ValueError('Invalid new species {} on site {}'.format(new_Z, site))

        # Update occupation vector itself
        self._occupations[sites] = species



"""
This module provides the Cluster class.
"""

from _icet import Cluster
from icet.core.lattice_site import LatticeSite
from icet.core.structure import Structure
__all__ = ['Cluster']


def _from_python(atoms, lattice_sites, sort=False, clusterIndex=-1):
    """
    Construct a cluster from atoms and python lattice sites
    """

    structure = Structure.from_atoms(atoms)

    lattice_sites_cpp = [
        LatticeSite(ls.index, ls.unitcell_offset) for ls in lattice_sites]

    return Cluster(structure, lattice_sites_cpp, sort, clusterIndex)


Cluster.from_python = _from_python



from copy import deepcopy
from icet.core.structure import Structure
from typing import List
from ase import Atoms
import copy
from itertools import product
from string import ascii_uppercase
import numpy as np
from icet.tools.geometry import chemical_symbols_to_numbers


class Sublattice:
    """
    This class stores and provides information about a specific
    sublattice. A sublattice is always supercell specific since
    it contains lattice indices.

    Parameters
    ----------
    chemical_symbols
        the allowed species on this sublattice
    indices
        the lattice indices the sublattice consists of

    """

    def __init__(self, chemical_symbols: List[str], indices: List[int], symbol: str):
        self._chemical_symbols = chemical_symbols
        self._indices = indices
        self._symbol = symbol
        self._numbers = chemical_symbols_to_numbers(chemical_symbols)

    @property
    def chemical_symbols(self):
        return copy.deepcopy(self._chemical_symbols)

    @property
    def atomic_numbers(self):
        return self._numbers.copy()

    @property
    def indices(self):
        return self._indices.copy()

    @property
    def symbol(self):
        """Symbol representation of sublattice, i.e. A, B, C, etc.."""
        return self._symbol


class Sublattices:
    """
    This class stores and provides information about the sublattices
    of a structure.

    Parameters
    ----------
    allowed_species
        list of the allowed species on each site of the primitve
        structure. For example this can be the chemical_symbols from
        a cluster space
    primitive_structure
        the primitive structure the allowed species reference to
    structure
        the structure that the sublattices will be based on
    """

    def __init__(self, allowed_species: List[List[str]], primitive_structure: Atoms,
                 structure: Atoms):
        self._structure = structure
        # sorted unique sites, this basically decides A, B, C... sublattices
        active_lattices = sorted(set([tuple(sorted(symbols))
                                      for symbols in allowed_species if len(symbols) > 1]))
        inactive_lattices = sorted(
            set([tuple(sorted(symbols)) for symbols in allowed_species if len(symbols) == 1]))
        self._allowed_species = active_lattices + inactive_lattices

        n = int(np.sqrt(len(self._allowed_species))) + 1
        symbols = [''.join(p) for r in range(1, n+1) for p in product(ascii_uppercase, repeat=r)]

        cpp_prim_structure = Structure.from_atoms(primitive_structure)
        self._sublattices = []
        sublattice_to_indices = [[] for _ in range(len(self._allowed_species))]
        for index, position in enumerate(structure.get_positions()):
            lattice_site = cpp_prim_structure.find_lattice_site_by_position(position)

            # Get allowed species on this site
            species = allowed_species[lattice_site.index]

            # Get what sublattice those species correspond to
            sublattice = self._allowed_species.index(tuple(sorted(species)))

            sublattice_to_indices[sublattice].append(index)

        for symbol, species, indices in zip(symbols, self._allowed_species, sublattice_to_indices):
            sublattice = Sublattice(chemical_symbols=species, indices=indices, symbol=symbol)
            self._sublattices.append(sublattice)

        # Map lattice index to sublattice index
        self._index_to_sublattice = {}
        for k, sublattice in enumerate(self):
            for index in sublattice.indices:
                self._index_to_sublattice[index] = k

    def __getitem__(self, key: int) -> Sublattice:
        """ Returns a sublattice according to key. """
        return self._sublattices[key]

    def __len__(self):
        """ Returns number of sublattices. """
        return len(self._sublattices)

    def get_sublattice_index(self, index: int) -> int:
        """ Returns the index of the sublattice the symbol
        or index in the structure belongs to.

        Parameters
        ----------
        index
            index of site in the structure
        """
        return self._index_to_sublattice[index]

    @property
    def allowed_species(self) -> List[List[str]]:
        """Lists of the allowed species on each sublattice, in order.
        """
        return deepcopy(self._allowed_species)

    def get_sublattice_sites(self, index: int) -> List[int]:
        """Returns the sites that belong to the sublattice with the
        corresponding index.

        Parameters
        ----------
        index
            index of the sublattice
        """
        return self[index].indices

    def get_allowed_symbols_on_site(self, index: int) -> List[str]:
        """Returns the allowed symbols on the site.

        Parameters
        -----------
        index
            lattice site index
        """
        return self[self._index_to_sublattice[index]].chemical_symbols

    def get_allowed_numbers_on_site(self, index: int) -> List[int]:
        """Returns the allowed atomic numbers on the site.

        Parameters
        -----------
        index
            lattice site index
        """
        return self[self._index_to_sublattice[index]].atomic_numbers

    @property
    def active_sublattices(self) -> List[Sublattice]:
        """ List of the active sublattices. """
        return [sl for sl in self if len(sl.chemical_symbols) > 1]

    @property
    def inactive_sublattices(self) -> List[Sublattice]:
        """ List of the active sublattices. """
        return [sl for sl in self if len(sl.chemical_symbols) == 1]

    def assert_occupation_is_allowed(self, chemical_symbols: List[str]):
        """Asserts that the current occupation obeys the sublattices."""
        if len(chemical_symbols) != len(self._structure):
            raise ValueError("len of input chemical symbols ({}) do not match len of supercell"
                             " ({})".format(len(chemical_symbols), len(self._structure)))
        for sl in self:
            for i in sl.indices:
                if not chemical_symbols[i] in sl.chemical_symbols:
                    msg = ('Occupations of structure not compatible with the sublattice.'
                           ' Site {} with occupation {} not allowed on sublattice {}'
                           .format(i, chemical_symbols[i], sl.chemical_symbols))
                    raise ValueError(msg)



""" Base data container class. """

import getpass
import json
import numbers
import tarfile
import tempfile
import socket

from collections import OrderedDict
from datetime import datetime
from typing import BinaryIO, Dict, List, TextIO, Tuple, Union

import numpy as np
import pandas as pd

from ase import Atoms
from ase.io import read as ase_read, Trajectory
from ase.io import write as ase_write
from icet import __version__ as icet_version
from ..observers.base_observer import BaseObserver


class Int64Encoder(json.JSONEncoder):

    def default(self, obj):
        if isinstance(obj, np.int64):
            return int(obj)
        return json.JSONEncoder.default(self, obj)


class BaseDataContainer:
    """
    Base data container for storing information concerned with
    Monte Carlo simulations performed with mchammer.

    Parameters
    ----------
    structure : ase.Atoms
        reference atomic structure associated with the data container

    ensemble_parameters : dict
        parameters associated with the underlying ensemble

    metadata : dict
        metadata associated with the data container
    """

    def __init__(self, structure: Atoms,
                 ensemble_parameters: dict,
                 metadata: dict = OrderedDict()):
        """
        Initializes a BaseDataContainer object.
        """
        if not isinstance(structure, Atoms):
            raise TypeError('structure is not an ASE Atoms object')

        self.structure = structure.copy()
        self._ensemble_parameters = ensemble_parameters
        self._metadata = metadata
        self._add_default_metadata()
        self._last_state = {}

        self._observables = set()
        self._data_list = []

    def append(self, mctrial: int,
               record: Dict[str, Union[int, float, list]]):
        """
        Appends data to data container.

        Parameters
        ----------
        mctrial
            current Monte Carlo trial step
        record
            dictionary of tag-value pairs representing observations

        Raises
        ------
        TypeError
            if input parameters have the wrong type

        """
        if not isinstance(mctrial, numbers.Integral):
            raise TypeError('mctrial has the wrong type: {}'.format(type(mctrial)))

        if self._data_list:
            if self._data_list[-1]['mctrial'] > mctrial:
                raise ValueError('mctrial values should be given in ascending'
                                 ' order. This error can for example occur'
                                 ' when trying to append to an existing data'
                                 ' container after having reset the time step.'
                                 ' Note that the latter happens automatically'
                                 ' when initializing a new ensemble.')

        if not isinstance(record, dict):
            raise TypeError('record has the wrong type: {}'.format(type(record)))

        for tag in record.keys():
            self._observables.add(tag)

        row_data = OrderedDict()
        row_data['mctrial'] = mctrial
        row_data.update(record)
        self._data_list.append(row_data)

    def _update_last_state(self, last_step: int, occupations: List[int],
                           accepted_trials: int, random_state: tuple):
        """Updates last state of the simulation: last step, occupation vector
        and number of accepted trial steps.

        Parameters
        ----------
        last_step
            last trial step
        occupations
            occupation vector observed during the last trial step
        accepted_trial
            number of current accepted trial steps
        random_state
            tuple representing the last state of the random generator
        """
        self._last_state['last_step'] = last_step
        self._last_state['occupations'] = occupations
        self._last_state['accepted_trials'] = accepted_trials
        self._last_state['random_state'] = random_state

    def apply_observer(self, observer: BaseObserver):
        """ Adds observer data from observer to data container.

        The observer will only be run for the mctrials for which the
        trajectory have been saved.

        The interval of the observer is ignored.

        Parameters
        ----------
        observer
            observer to be used
        """
        for row_data in self._data_list:
            if 'occupations' in row_data:
                structure = self.structure.copy()
                structure.numbers = row_data['occupations']
                record = dict()
                if observer.return_type is dict:
                    for key, value in observer.get_observable(structure).items():
                        record[key] = value
                else:
                    record[observer.tag] = observer.get_observable(structure)
                row_data.update(record)
                self._observables.add(observer.tag)

    def get_data(self, *tags,
                 start: int = None,
                 stop: int = None,
                 interval: int = 1,
                 fill_method: str = 'skip_none',
                 apply_to: List[str] = None) \
            -> Union[np.ndarray, List[Atoms], Tuple[np.ndarray, List[Atoms]]]:
        """Returns the accumulated data for the requested observables,
        including configurations stored in the data container. The latter
        can be achieved by including 'trajectory' as a tag.

        Parameters
        ----------
        tags
            tuples of the requested properties

        start
            minimum value of trial step to consider; by default the
            smallest value in the mctrial column will be used.

        stop
            maximum value of trial step to consider; by default the
            largest value in the mctrial column will be used.

        interval
            increment for mctrial; by default the smallest available
            interval will be used.

        fill_method : {'skip_none', 'fill_backward', 'fill_forward',
                       'linear_interpolate', None}
            method employed for dealing with missing values; by default
            uses 'skip_none'.

        apply_to
            tags of columns for which fill_method will be employed;
            by default parse all columns with fill_method.

        Raises
        ------
        ValueError
            if tags is empty
        ValueError
            if observables are requested that are not in data container
        ValueError
            if fill method is unknown
        ValueError
            if trajectory is requested and fill method is not skip_none

        Examples
        --------
        The following lines illustrate how to use the `get_data` method
        for extracting data from the trajectory::

            # obtain a list of all values of the potential represented by
            # the cluster expansion along the trajectory
            p = dc.get_data('potential')

            # as above but this time the MC trial step and the temperature
            # are included as well
            s, p, t = dc.get_data('mctrial', 'potential', 'temperature')

            # obtain configurations along the trajectory along with
            # their potential
            p, confs = dc.get_data('potential', 'trajectory')

        """
        fill_methods = ['skip_none',
                        'fill_backward',
                        'fill_forward',
                        'linear_interpolate']

        if len(tags) == 0:
            raise TypeError('Missing tags argument')

        if 'trajectory' in tags:
            if fill_method != 'skip_none':
                raise ValueError('Only skip_none fill method is avaliable'
                                 ' when trajectory is requested')

            new_tags = tuple(['occupations' if tag == 'trajectory' else tag for tag in tags])
            return self._get_trajectory(*new_tags, start=start, stop=stop, interval=interval)

        for tag in tags:
            if tag == 'mctrial':
                continue
            if tag not in self.observables:
                raise ValueError('No observable named {} in data container'.format(tag))

        mctrials = [row_dict['mctrial'] for row_dict in self._data_list]
        data = pd.DataFrame.from_records(self._data_list, index=mctrials, columns=tags)
        if start is None and stop is None:
            data = data.loc[::interval, tags].copy()
        else:
            # slice and pass a copy to avoid slowing down dropna method below
            if start is None:
                data = data.loc[:stop:interval, tags].copy()
            elif stop is None:
                data = data.loc[start::interval, tags].copy()
            else:
                data = data.loc[start:stop:interval, tags].copy()

        if fill_method is not None:
            if fill_method not in fill_methods:
                raise ValueError('Unknown fill method: {}'
                                 .format(fill_method))

            if apply_to is None:
                apply_to = tags

            # retrieve only valid observations
            if fill_method == 'skip_none':
                data.dropna(inplace=True, subset=apply_to)

            else:
                # if requested, drop NaN values in columns
                subset = [tag for tag in tags if tag not in apply_to]
                data.dropna(inplace=True, subset=subset)

                # fill NaN with the next valid observation
                if fill_method == 'fill_backward':
                    data.fillna(method='bfill', inplace=True)

                # fill NaN with the last valid observation
                elif fill_method == 'fill_forward':
                    data.fillna(method='ffill', inplace=True)

                # fill NaN with the linear interpolation of the last and
                # next valid observations
                elif fill_method == 'linear_interpolate':
                    data.interpolate(limit_area='inside', inplace=True)

                # drop any left-over nan values
                data.dropna(inplace=True)

        data_list = []
        for tag in tags:
            # convert NaN to None
            data_list.append(np.array([None if np.isnan(x).any() else x for x in data[tag]]))
        if len(tags) > 1:
            # return a tuple if more than one tag is given
            return tuple(data_list)
        else:
            # return a list if only one tag is given
            return data_list[0]

    @property
    def data(self) -> pd.DataFrame:
        """ pandas data frame (see :class:`pandas.DataFrame`) """
        if self._data_list:
            df = pd.DataFrame.from_records(self._data_list, index='mctrial',
                                           exclude=['occupations'])
            df.dropna(axis='index', how='all', inplace=True)
            df.reset_index(inplace=True)
            return df
        else:
            return pd.DataFrame()

    @property
    def ensemble_parameters(self) -> dict:
        """ parameters associated with Monte Carlo simulation """
        return self._ensemble_parameters.copy()

    @property
    def observables(self) -> List[str]:
        """ observable names """
        return list(self._observables)

    @property
    def metadata(self) -> dict:
        """ metadata associated with data container """
        return self._metadata

    @property
    def last_state(self) -> Dict[str, Union[int, List[int]]]:
        """ last state to be used to restart Monte Carlo simulation """
        return self._last_state

    def get_number_of_entries(self, tag: str = None) -> int:
        """
        Returns the total number of entries with the given observable tag.

        Parameters
        ----------
        tag
            name of observable; by default the total number of rows in the
            data frame will be returned.

        Raises
        ------
        ValueError
            if observable is requested that is not in data container
        """
        data = pd.DataFrame.from_records(self._data_list)
        if tag is None:
            return len(data)
        else:
            if tag not in data:
                raise ValueError('No observable named {} in data container'.format(tag))
            return data[tag].count()

    def get_trajectory(self, start: int = None, stop: int = None, interval: int = 1) -> List[Atoms]:
        """ Returns trajectory as a list of ASE Atoms objects.

        Parameters
        ----------
        start
            minimum value of trial step to consider; by default the
            smallest value in the mctrial column will be used.
        stop
            maximum value of trial step to consider; by default the
            largest value in the mctrial column will be used.
        interval
            increment for mctrial; by default the smallest available
            interval will be used.
        """
        return self.get_data('trajectory', start=start, stop=stop, interval=interval)

    def _get_trajectory(self, *tags, start: int = None, stop: int = None, interval: int = 1) \
            -> Union[List[Atoms], Tuple[List[Atoms], np.ndarray]]:
        """
        Returns a trajectory in the form of a list of ASE Atoms
        along with the corresponding values of the mctrial and/or scalar
        properties upon request.
        Configurations with non properties will be
        skipped in the trajectory if the property is requested.

        Parameters
        ----------
        start
            minimum value of trial step to consider; by default the
            smallest value in the mctrial column will be used.
        stop
            maximum value of trial step to consider; by default the
            largest value in the mctrial column will be used.
        interval
            increment for mctrial; by default the smallest available
            interval will be used.
        """

        if 'occupations' in tags:
            new_tags = tags
        else:
            new_tags = ('occupations', ) + tags

        data = self.get_data(*new_tags, start=start, stop=stop, interval=interval)

        if len(new_tags) > 1:
            data_list = list(data)
        else:
            data_list = [data]

        tag_list = list(new_tags)
        structure_list = []
        for tag, data_row in zip(tag_list, data_list):
            if tag == 'occupations':
                ind = tag_list.index('occupations')
                for occupation_vector in data_row:
                    structure = self.structure.copy()
                    structure.numbers = occupation_vector
                    structure_list.append(structure)
                data_list[ind] = structure_list

        if len(data_list) > 1:
            return tuple(data_list)
        else:
            return data_list[0]

    def write_trajectory(self, outfile: Union[str, BinaryIO, TextIO]) -> None:
        """Writes the configurations along the trajectory to file in ASE
        trajectory format.  The file also includes the respectives
        values of the potential for each configuration. If the file
        exists the trajectory will be appended. The ASE `convert`
        command can be used to convert the trajectory file to other
        formats. The ASE `gui` can be used to visualize the
        trajectory.

        Parameters
        ----------
        outfile
            output file name or file object
        """
        structure_list, energies = self._get_trajectory('occupations', 'potential')
        traj = Trajectory(outfile, mode='a')
        for structure, energy in zip(structure_list, energies):
            traj.write(atoms=structure, energy=energy)
        traj.close()

    def write(self, outfile: Union[str, BinaryIO, TextIO]):
        """
        Writes BaseDataContainer object to file.

        Parameters
        ----------
        outfile
            file to which to write
        """
        self._metadata['date_last_backup'] = datetime.now().strftime('%Y-%m-%dT%H:%M:%S')

        # Save reference atomic structure
        reference_structure_file = tempfile.NamedTemporaryFile()
        ase_write(reference_structure_file.name, self.structure, format='json')

        # Save reference data
        data_container_type = str(self.__class__).split('.')[-1].replace("'>", '')
        reference_data = {'parameters': self._ensemble_parameters,
                          'metadata': self._metadata,
                          'last_state': self._last_state,
                          'data_container_type': data_container_type}

        reference_data_file = tempfile.NamedTemporaryFile()
        with open(reference_data_file.name, 'w') as handle:
            json.dump(reference_data, handle, cls=Int64Encoder)

        # Save runtime data
        runtime_data_file = tempfile.NamedTemporaryFile()
        np.savez_compressed(runtime_data_file, self._data_list)

        with tarfile.open(outfile, mode='w') as handle:
            handle.add(reference_structure_file.name, arcname='atoms')
            handle.add(reference_data_file.name, arcname='reference_data')
            handle.add(runtime_data_file.name, arcname='runtime_data')
        runtime_data_file.close()

    def _add_default_metadata(self):
        """Adds default metadata to metadata dict."""

        self._metadata['date_created'] = \
            datetime.now().strftime('%Y-%m-%dT%H:%M:%S')
        self._metadata['username'] = getpass.getuser()
        self._metadata['hostname'] = socket.gethostname()
        self._metadata['icet_version'] = icet_version

    def __str__(self):
        """ string representation of data container """
        width = 80
        s = []  # type: List
        s += ['{s:=^{n}}'.format(s=' Data Container ', n=width)]
        data_container_type = str(self.__class__).split('.')[-1].replace("'>", '')
        s += [' {:22}: {}'.format('data_container_type', data_container_type)]
        for key, value in self._last_state.items():
            if isinstance(value, int) or isinstance(value, float) or isinstance(value, str):
                s += [' {:22}: {}'.format(key, value)]
        for key, value in sorted(self._ensemble_parameters.items()):
            s += [' {:22}: {}'.format(key, value)]
        for key, value in sorted(self._metadata.items()):
            s += [' {:22}: {}'.format(key, value)]
        s += [' {:22}: {}'.format('columns_in_data', self.data.columns.tolist())]
        s += [' {:22}: {}'.format('n_rows_in_data', len(self.data))]
        s += [''.center(width, '=')]
        return '\n'.join(s)

    @classmethod
    # todo: cls and the return should be type hinted as BaseDataContainer.
    # Unfortunately, this requires from __future__ import annotations, which
    # in turn requires Python 3.8.
    def read(cls,
             infile: Union[str, BinaryIO, TextIO],
             old_format: bool = False):
        """Reads data container from file.

        Parameters
        ----------
        infile
            file from which to read
        old_format
            If true use old json format to read runtime data; default to false

        Raises
        ------
        FileNotFoundError
            if file is not found (str)
        ValueError
            if file is of incorrect type (not a tarball)
        """
        if isinstance(infile, str):
            filename = infile
        else:
            filename = infile.name

        if not tarfile.is_tarfile(filename):
            raise TypeError('{} is not a tar file'.format(filename))

        reference_structure_file = tempfile.NamedTemporaryFile()
        reference_data_file = tempfile.NamedTemporaryFile()
        runtime_data_file = tempfile.NamedTemporaryFile()

        with tarfile.open(mode='r', name=filename) as tar_file:
            # file with structures
            reference_structure_file.write(tar_file.extractfile('atoms').read())

            reference_structure_file.seek(0)
            structure = ase_read(reference_structure_file.name, format='json')

            # file with reference data
            reference_data_file.write(tar_file.extractfile('reference_data').read())
            reference_data_file.seek(0)
            with open(reference_data_file.name, encoding='utf-8') as fd:
                reference_data = json.load(fd)

            # init DataContainer
            dc = cls(structure=structure,
                     ensemble_parameters=reference_data['parameters'])

            # overwrite metadata
            dc._metadata = reference_data['metadata']

            for tag, value in reference_data['last_state'].items():
                if tag == 'random_state':
                    value = tuple(tuple(x) if isinstance(x, list) else x for x in value)
                if tag in ['histogram', 'entropy']:
                    # the following accounts for the fact that the keys of dicts are converted to
                    # str when writing to json and have to converted back into numerical values
                    dc._last_state[tag] = {int(k): v for k, v in value.items()}
                else:
                    dc._last_state[tag] = value

            # add runtime data from file
            runtime_data_file.write(tar_file.extractfile('runtime_data').read())
            runtime_data_file.seek(0)
            if old_format:
                runtime_data = pd.read_json(runtime_data_file)
                data = runtime_data.sort_index(ascending=True)
                dc._data_list = data.T.apply(lambda x: x.dropna().to_dict()).tolist()
            else:
                dc._data_list = np.load(runtime_data_file, allow_pickle=True)['arr_0'].tolist()

        dc._observables = set([key for data in dc._data_list for key in data])
        dc._observables = dc._observables - {'mctrial'}

        return dc



"""Definition of the canonical annealing class."""

import numpy as np

from ase import Atoms
from ase.units import kB
from typing import Dict, List

from .. import DataContainer
from ..calculators.base_calculator import BaseCalculator
from .thermodynamic_base_ensemble import ThermodynamicBaseEnsemble


class CanonicalAnnealing(ThermodynamicBaseEnsemble):
    """Instances of this class allow one to carry out simulated annealing
    in the canonical ensemble, i.e. the temperature is varied in
    pre-defined fashion while the composition is kept fixed.  See
    :class:`mchammer.ensembles.CanonicalEnsemble` for more information
    about the standard canonical ensemble.

    The canonical annealing ensemble can be useful, for example, for
    finding ground states or generating low energy configurations.

    The temperature control scheme is selected via the
    ``cooling_function`` keyword argument, while the initial and final
    temperature are set via the ``T_start`` and ``T_stop`` arguments.
    Several pre-defined temperature control schemes are available
    including `linear` and `exponential`. In the latter case the
    temperature varies logarithmatically as a function of the MC step,
    emulating the exponential temperature dependence of the atomic
    exchange rate encountered in many materials.  It is also possible
    to provide a user defined cooling function via the keyword
    argument.  This function must comply with the following function
    header::

        def cooling_function(step, T_start, T_stop, n_steps):
            T = ...  # compute temperature
            return T

    Here ``step`` refers to the current MC trial step.

    Parameters
    ----------
    structure : :class:`Atoms <ase.Atoms>`
        atomic configuration to be used in the Monte Carlo simulation;
        also defines the initial occupation vector
    calculator : :class:`BaseCalculator <mchammer.calculators.ClusterExpansionCalculator>`
        calculator to be used for calculating the potential changes
        that enter the evaluation of the Metropolis criterion
    T_start : float
        temperature from which the annealing is started
    T_stop : float
        final temperature for annealing
    n_steps : int
        number of steps to take in the annealing simulation
    cooling_function : str/function
        to use the predefined cooling functions provide a string
        `linear` or `exponential`, otherwise provide a function
    boltzmann_constant : float
        Boltzmann constant :math:`k_B` in appropriate
        units, i.e. units that are consistent
        with the underlying cluster expansion
        and the temperature units [default: eV/K]
    user_tag : str
        human-readable tag for ensemble [default: None]
    random_seed : int
        seed for the random number generator used in the Monte Carlo
        simulation
    data_container : str
        name of file the data container associated with the ensemble
        will be written to; if the file exists it will be read, the
        data container will be appended, and the file will be
        updated/overwritten
    data_container_write_period : float
        period in units of seconds at which the data container is
        written to file; writing periodically to file provides both
        a way to examine the progress of the simulation and to back up
        the data [default: np.inf]
    ensemble_data_write_interval : int
        interval at which data is written to the data container; this
        includes for example the current value of the calculator
        (i.e. usually the energy) as well as ensembles specific fields
        such as temperature or the number of atoms of different species
    trajectory_write_interval : int
        interval at which the current occupation vector of the atomic
        configuration is written to the data container.
    sublattice_probabilities : List[float]
        probability for picking a sublattice when doing a random swap.
        This should be as long as the number of sublattices and should
        sum up to 1.

    """

    def __init__(self,
                 structure: Atoms,
                 calculator: BaseCalculator,
                 T_start: float,
                 T_stop: float,
                 n_steps: int,
                 cooling_function: str = 'exponential',
                 user_tag: str = None,
                 boltzmann_constant: float = kB,
                 random_seed: int = None,
                 data_container: str = None,
                 data_container_write_period: float = np.inf,
                 ensemble_data_write_interval: int = None,
                 trajectory_write_interval: int = None,
                 sublattice_probabilities: List[float] = None) -> None:

        self._ensemble_parameters = dict(n_steps=n_steps)

        # add species count to ensemble parameters
        for sl in calculator.sublattices:
            for symbol in sl.chemical_symbols:
                key = 'n_atoms_{}'.format(symbol)
                count = structure.get_chemical_symbols().count(symbol)
                self._ensemble_parameters[key] = count

        super().__init__(
            structure=structure, calculator=calculator, user_tag=user_tag,
            random_seed=random_seed,
            data_container=data_container,
            data_container_class=DataContainer,
            data_container_write_period=data_container_write_period,
            ensemble_data_write_interval=ensemble_data_write_interval,
            trajectory_write_interval=trajectory_write_interval,
            boltzmann_constant=boltzmann_constant)

        self._temperature = T_start
        self._T_start = T_start
        self._T_stop = T_stop
        self._n_steps = n_steps

        self._ground_state_candidate = self.configuration.structure
        self._ground_state_candidate_potential = self.calculator.calculate_total(
            occupations=self.configuration.occupations)

        # setup cooling function
        if isinstance(cooling_function, str):
            available = sorted(available_cooling_functions.keys())
            if cooling_function not in available:
                raise ValueError(
                    'Select from the available cooling_functions {}'.format(available))
            self._cooling_function = available_cooling_functions[cooling_function]
        elif callable(cooling_function):
            self._cooling_function = cooling_function
        else:
            raise TypeError('cooling_function must be either str or a function')

        if sublattice_probabilities is None:
            self._swap_sublattice_probabilities = self._get_swap_sublattice_probabilities()
        else:
            self._swap_sublattice_probabilities = sublattice_probabilities

    @property
    def temperature(self) -> float:
        """ Current temperature """
        return self._temperature

    @property
    def T_start(self) -> float:
        """ Starting temperature """
        return self._T_start

    @property
    def T_stop(self) -> float:
        """ Starting temperature """
        return self._T_stop

    @property
    def n_steps(self) -> int:
        """ Number of steps to carry out """
        return self._n_steps

    @property
    def estimated_ground_state(self):
        """ Structure with lowest observed potential during run """
        return self._ground_state_candidate.copy()

    @property
    def estimated_ground_state_potential(self):
        """ Lowest observed potential during run """
        return self._ground_state_candidate_potential

    def run(self):
        """ Runs the annealing. """
        if self.step >= self.n_steps:
            raise Exception('Annealing has already finished')
        super().run(self.n_steps - self.step)

    def _do_trial_step(self) -> int:
        """ Carries out one Monte Carlo trial step. """
        self._temperature = self._cooling_function(
            self.step, self.T_start, self.T_stop, self.n_steps)
        sublattice_index = self.get_random_sublattice_index(self._swap_sublattice_probabilities)
        return self.do_canonical_swap(sublattice_index=sublattice_index)

    def _get_ensemble_data(self) -> Dict:
        """Returns the data associated with the ensemble. For the
        CanonicalAnnealing this specifically includes the temperature.
        """
        data = super()._get_ensemble_data()
        data['temperature'] = self.temperature
        if data['potential'] < self._ground_state_candidate_potential:
            self._ground_state_candidate_potential = data['potential']
            self._ground_state_candidate = self.configuration.structure
        return data


def _cooling_linear(step, T_start, T_stop, n_steps):
    return T_start + (T_stop-T_start) * step / (n_steps - 1)


def _cooling_exponential(step, T_start, T_stop, n_steps):
    return T_start - (T_start - T_stop) * np.log(step+1) / np.log(n_steps)


available_cooling_functions = dict(linear=_cooling_linear, exponential=_cooling_exponential)



# -*- coding: utf-8 -*-

from .wang_landau_data_container import (get_average_cluster_vectors_wl,
                                         get_average_observables_wl,
                                         get_density_of_states_wl)

__all__ = ['get_average_cluster_vectors_wl',
           'get_average_observables_wl',
           'get_density_of_states_wl']
